// Requires overlay.js and jquery

// Highlight DOM elements with an overlayed box, similar to Webkit's inspector.
// Creates an absolute-positioned div that is translated & scaled following
// mousemove events. Holds a pointer to target DOM element.
function Highlighter() {
  var target = null;
  var enabled = false;
  var then = Date.now();
  var $box = $("<div class='adblock-highlight-node'></div>");
  $box.appendTo('body');

  function handler(e) {
    var offset;
    var el = e.target;
    var now = Date.now();
    if (now - then < 25) {
      return;
    }

    then = now;
    if (el === $box[0]) {
      $box.hide();
      el = document.elementFromPoint(e.clientX, e.clientY);
    }

    if (el === target) {
      $box.show();
      return;
    }

    if (el === document.body || el.className === 'adblock-killme-overlay') {
      $box.hide();
      return;
    }

    $el = $(el);
    target = $el[0];
    offset = $el.offset();
    $box.css({
      height: $el.outerHeight(),
      width: $el.outerWidth(),
      left: offset.left,
      top: offset.top,
    });
    $box.show();
  }

  this.getCurrentNode = function (el) {
    return el === $box[0] ? target : el;
  };

  this.enable = function () {
    if ($box && !enabled) {
      $('body').bind('mousemove', handler);
    }

    enabled = true;
  };

  this.disable = function () {
    if ($box && enabled) {
      $box.hide();
      $('body').unbind('mousemove', handler);
    }

    enabled = false;
  };

  this.destroy = function () {
    this.disable();
    if ($box) {
      $box.remove();
      $box = null;
    }
  };
}

// Class that watches the whole page for a click, including iframes and
// objects.  Shows a modal while doing so.
function ClickWatcher() {
  this._callbacks = { cancel: [], click: [] };
  this._clickedElement = null;
  this._highlighter = new Highlighter();
}

ClickWatcher.prototype.cancel = function (callback) {
  this._callbacks.cancel.push(callback);
};

ClickWatcher.prototype.click = function (callback) {
  this._callbacks.click.push(callback);
};

ClickWatcher.prototype._fire = function (eventName, arg) {
  var callbacks = this._callbacks[eventName];
  for (var i = 0; i < callbacks.length; i++)
    callbacks[i](arg);
};

ClickWatcher.prototype.show = function () {
  var _this = this;
  var wait = $('<div style="white-space: pre-wrap;" ></div>').
    text(translate('findingads')).
    dialog({
      classes: {
        "ui-dialog": 'adblock-blacklist-dialog'
      },
      closeText: "",
      position: { my: "left top", at: "left+10% top+10%", of: window },
      height: 120,
      minHeight: 120,
      title: translate('blockanadtitle'),
    });
  changeTextDirection($('body .adblock-blacklist-dialog'));

  // setTimeout to give 'wait' a chance to display
  window.setTimeout(function () {
    _this._ui = _this._build_ui();
    wait.dialog('close');
    wait.remove();
    _this._ui.dialog('open');
    _this._highlighter.enable();
  }, 10);
};

// Called externally to close ClickWatcher.  Doesn't cause any events to
// fire.
ClickWatcher.prototype.close = function () {

  // Delete our event listeners so we don't fire any cancel events
  this._callbacks.cancel = [];
  if (this._ui) {
    this._ui.dialog('close');
  }
};

// The dialog is closing, either because the user clicked cancel, or the
// close button, or because they clicked an item.
ClickWatcher.prototype._onClose = function () {
  if (this._clickedElement == null) {

    // User clicked Cancel button or X
    this._fire('cancel');
  } else {

    // User clicked a page item
    this._fire('click', this._clickedElement);
  }

  this._highlighter.destroy();
};

ClickWatcher.prototype._build_ui = function () {
  var _this = this;

  function clickCatchThis() {
    return clickCatch(this);
  }

  function clickCatch(element) {
    _this._clickedElement = _this._highlighter.getCurrentNode(element);

    //_this._clickedElement = element;
    _this._ui.dialog('close');
    return false;
  }

  // Most things can be blacklisted with a simple click handler.
  $('body').on('click', '.adblock-killme-overlay, .adblock-highlight-node',
    clickCatchThis);

  // Since iframes that will get clicked will almost always be an entire
  // ad, and I *really* don't want to figure out inter-frame communication
  // so that the blacklist UI's slider works between multiple layers of
  // iframes... just overlay iframes and treat them as a giant object.
    $("object,embed,iframe,[onclick]:empty").
      each(function(i, el) {
        // Don't add overlay's for hidden elements
        if (el.style &&
            el.style.display === "none")
        {
          return;
        }
        var killmeOverlay = new Overlay({
          dom_element: el,
          click_handler: clickCatch
        });
        killmeOverlay.display();
  });

  var btn = {};
  btn[translate('buttoncancel')] = function () {
    $('.adblock-ui-stylesheet').remove();
    page.dialog('close');
  };

  var page = $('<div></div>').
    text(translate('clickthead')).
    append('<br/><br/>').
    dialog({
      classes: {
        "ui-dialog": 'adblock-blacklist-dialog'
      },
      closeText: "",
      position: { my: "left top", at: "left+10% top+10%", of: window },
      width: 400,
      minHeight: 125,
      autoOpen: false,
      title: translate('blockanadtitle'),
      buttons: btn,
      close: function () {
        $('body').off('click',
          '.adblock-killme-overlay, .adblock-highlight-node', clickCatchThis);
        Overlay.removeAll();
        _this._onClose();
        page.remove();
      },

      drag: function () {
        _this._highlighter.disable();
      },
    });
  page.dialog('widget').
    bind('mouseenter', function () {
      _this._highlighter.disable();
    }).
    bind('mouseleave', function () {
      _this._highlighter.enable();
    });

  changeTextDirection($('body .adblock-blacklist-dialog'));
  return page;
};

//# sourceURL=/uiscripts/blacklisting/clickwatcher.js
